library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.FfpgPkg.all;

entity ClkRfDomain is
    port (
        -- main signals
        Clk_ik: in std_logic;
        Reset_ir: in std_logic;
        -- WB interface
        Overflow_ib16: in unsigned(15 downto 0);
        TriggerLatency_ib16: in unsigned(15 downto 0);
        Mode_i: in t_Mode;
        ModeLoad_i: in std_logic;
        SetMemAddress_ob11: out unsigned(10 downto 0);
        SetMemData_ib32: in unsigned(31 downto 0);
        SetMemReadStrobe_o: out std_logic;
        ResMemAddress_ob11: out unsigned(10 downto 0);
        ResMemData_ib32: in unsigned(31 downto 0);
        ResMemReadStrobe_o: out std_logic;
        Running_o: out std_logic;
        -- FMC interface
        Trigger_i: in std_logic;
        SetStream_o: out std_logic;
        ResetStream_o: out std_logic
    );
end entity;

architecture syn of ClkRfDomain is

    constant c_MemLatency: integer := 2;

    signal LastStreamPosition: unsigned(Overflow_ib16'range);
    signal TriggerLatencyPlusTwo: unsigned(TriggerLatency_ib16'range);

    signal GenerationEnable: std_logic;
    signal OutputEnable: std_logic;
    
    signal AddressEnableCounterReset, AddressEnableCounterOverflow: std_logic;
    signal AddressEnableCounterSetValue: unsigned(TriggerLatency_ib16'range);
    
    signal AddressCounterReset: std_logic;
    signal AddressCounterSetValue: unsigned(TriggerLatency_ib16'range);
    signal AddressCounterValue: unsigned(SetMemAddress_ob11'range);
    
    signal StreamReset, StreamResetOrReset: std_logic;
    signal StreamPosition: unsigned(Overflow_ib16'range);

    signal BitCounterOverflow, LoadShiftRegister: std_logic;
    
    signal SetStream, ResetStream: std_logic;
    
begin

    -- for better timing
    pInputReg: process(Clk_ik) is begin
        if rising_edge(Clk_ik) then
            LastStreamPosition <= Overflow_ib16 - 1;
            TriggerLatencyPlusTwo <= TriggerLatency_ib16 + 2;
                -- +1 because of ???
                -- +1 because of output registers on Set/ResetStream_o
        end if;
    end process;

    cFsm: entity work.Fsm(syn)
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Mode_i => Mode_i,
            ModeLoad_i => ModeLoad_i,
            Trigger_i => Trigger_i,
            StreamReset_i => StreamReset,
            Running_o => Running_o,
            GenerationEnable_o => GenerationEnable,
            OutputEnable_o => OutputEnable
        );

    -- for better timing
    pAddressEnableCounterReset: process(Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if StreamPosition = (LastStreamPosition - c_MemLatency) - 1 then
                AddressEnableCounterReset <= '1';
            else
                AddressEnableCounterReset <= '0';
            end if;
        end if;
    end process;

    AddressEnableCounterSetValue <= TriggerLatencyPlusTwo + c_MemLatency;
    cAddressEnableCounter: entity work.Counter(syn)
        generic map (
            g_Width => 5
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => AddressEnableCounterReset,
            Enable_i => GenerationEnable,
            Set_i => Trigger_i,
            SetValue_ib => AddressEnableCounterSetValue(4 downto 0),
            Overflow_o => AddressEnableCounterOverflow,
            Value_ob => open
        );

    AddressCounterReset <=
        '1' when StreamPosition = (LastStreamPosition - c_MemLatency) else
        Reset_ir;
    AddressCounterSetValue <= (TriggerLatencyPlusTwo + c_MemLatency);
    cAddressCounter: entity work.Counter(syn)
        generic map (
            g_Width => SetMemAddress_ob11'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => AddressCounterReset,
            Enable_i => AddressEnableCounterOverflow,
            Set_i => Trigger_i,
            SetValue_ib => AddressCounterSetValue(15 downto 5),
            Overflow_o => open,
            Value_ob => AddressCounterValue
        );
    
    SetMemAddress_ob11 <= AddressCounterValue;
    ResMemAddress_ob11 <= AddressCounterValue;
    
    SetMemReadStrobe_o <= '0'; -- this is not connected anywhere!!
    ResMemReadStrobe_o <= '0'; -- this is not connected anywhere!!
    
    -- for better timing
    pStreamResetReg: process(Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if StreamPosition = Overflow_ib16 - 2 then
                StreamReset <= '1';
            else
                StreamReset <= '0';
            end if;
        end if;
    end process;
    
    StreamResetOrReset <= StreamReset or Reset_ir;

    cStreamCounter: entity work.Counter(syn)
        generic map (
            g_Width => Overflow_ib16'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => StreamResetOrReset,
            Enable_i => GenerationEnable,
            Set_i => Trigger_i,
            SetValue_ib => TriggerLatencyPlusTwo,
            Overflow_o => open,
            Value_ob => StreamPosition
        );
    
    cBitCounter: entity work.Counter(syn)
        generic map (
            g_Width => 5
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => StreamResetOrReset,
            Enable_i => GenerationEnable,
            Set_i => Trigger_i,
            SetValue_ib => TriggerLatencyPlusTwo(4 downto 0),
            Overflow_o => BitCounterOverflow,
            Value_ob => open
        );
    
    LoadShiftRegister <= BitCounterOverflow or StreamReset;
    
    cShiftRegisterSet: entity work.ShiftRegister(right)
        generic map (
            g_Width => SetMemData_ib32'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Data_ib => SetMemData_ib32,
            Data_ob => open,
            Shift_i => '0',
            Shift_o => SetStream,
            ShiftEnable_i => GenerationEnable,
            LoadEnable_i => LoadShiftRegister
        );
    
    cShiftRegisterReset: entity work.ShiftRegister(right)
        generic map (
            g_Width => ResMemData_ib32'length
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Data_ib => ResMemData_ib32,
            Data_ob => open,
            Shift_i => '0',
            Shift_o => ResetStream,
            ShiftEnable_i => GenerationEnable,
            LoadEnable_i => LoadShiftRegister
        );
    
    pOutputRegs: process (Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if OutputEnable = '1' then
                SetStream_o <= SetStream;
                ResetStream_o <= ResetStream;
            else
                SetStream_o <= '0';
                ResetStream_o <= '0';
            end if;
        end if;
    end process;
    
end architecture;