#!/usr/bin/env python

import sys
sys.path.append('/usr/local/encore')
import encoreio

class ConColors:
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m' # better readable with BOLD
    MAGENTA = '\033[35m'
    CYAN = '\033[36m'
    WHITE = '\033[37m'
    NONE = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    
m = encoreio.Module.get_instance('fmc_fpg', 0)

WB_DEBUG = False

###################################################################
## WishBone access overlay
###################################################################

def WbRead(register):
    result = m.read(register)[0]
    if WB_DEBUG:
        print(ConColors.MAGENTA+'WB_READ:'+ConColors.NONE+' ['+register+'] = '+hex(result))
    return result

def WbWrite(register, data):
    if WB_DEBUG:
        print(ConColors.CYAN+'WB_WRITE:'+ConColors.NONE+' ['+register+'] <- '+hex(data))
    m.write(register, (data,))

def WbSetBits(register, mask, bits):
    value = WbRead(register)
    value &= (~mask)
    value |= (bits & mask)
    WbWrite(register, value)

def WbReadMulti(register, start = 0, to = -1):
    result = m.read(register, start=start, to=to)
    if WB_DEBUG:
        print(ConColors.MAGENTA+'WB_READ_MULTI:'+ConColors.NONE+' ['+register+']('+str(start)+':'+str(to)+') = ' + ', '.join([hex(x) for x in result]))
    return result

def WbWriteMulti(register, data, start = 0, to = -1):
    if WB_DEBUG:
        print(ConColors.CYAN+'WB_WRITE_MULTI:'+ConColors.NONE+' ['+register+']('+str(start)+':'+str(to)+') <- ' + ', '.join([hex(x) for x in data]))
    m.write(register, data, start=start, to=to)

def WbSetBitsMulti(register, mask, bits, start = 0, to = -1):
    if len(mask) != len(bits):
        raise Exception('Different MASK and BITS length!')
    values = WbReadMulti(register, start, to)
    values = list(values)
    if len(values) != len(mask):
        raise Exception('Different length of MASK/BITS and actual data!');
    for i in range(len(values)):
        values[i] &= (~mask[i])
        values[i] |= (bits[i] & mask[i])
    WbWriteMulti(register, values, start, to)
    
###################################################################
## SPI functions
###################################################################

def SpiInit():
    WbWrite('spi_ss', 1)
    WbWrite('spi_divider', 2)
    WbWrite('spi_ctrl', 0x2618)

def SpiRead(register):
    inst = 0x800000 | ((register & 0xff) << 8)
    WbWrite('spi_tx_rx_0', inst)
    WbSetBits('spi_ctrl', 0x100, 0x100)
    return WbRead('spi_tx_rx_0')

def SpiWrite(register, data):
    inst = 0x000000 | ((register & 0xff) << 8) | (data & 0xff)
    WbWrite('spi_tx_rx_0', inst)
    WbSetBits('spi_ctrl', 0x100, 0x100)

###################################################################
## FFPG specific functions
###################################################################

def Ad9512Init():
    # power down unused output (OUT2)
    SpiWrite(0x3f, 0x03)
    # fine delay
    SpiWrite(0x34, 0x0) # use fine delay
    delay = 2 # 0-31
    SpiWrite(0x36, (delay&0x1f)<<1) # set fine delay
    # select CLK1 input, power down CLK2 input
    SpiWrite(0x45, 0x05)
    # output frequency 200 MHz (divide by 2)
    SpiWrite(0x4a, 0x00) # OUT0
    SpiWrite(0x4c, 0x00) # OUT1
    SpiWrite(0x4e, 0x00) # OUT2
    SpiWrite(0x50, 0x00) # OUT3
    SpiWrite(0x52, 0x00) # OUT4
    # phase 0
    SpiWrite(0x4b, 0x00) # OUT0
    SpiWrite(0x4d, 0x00) # OUT1
    SpiWrite(0x4f, 0x00) # OUT2
    SpiWrite(0x51, 0x00) # OUT3
    SpiWrite(0x53, 0x00) # OUT4
    # function pin as sync
    SpiWrite(0x58, 0x20)
    # confirm write
    SpiWrite(0x5a, 1)

# 0 - front panel clock
# 1 - FPGA loop clock
# 2 - oscilator
def SelectClock(clock):
    if clock == 0:
        # select CLK1 input, power down CLK2 input (AD9512)
        SpiWrite(0x45, 0x05)
        # confirm write
        SpiWrite(0x5a, 1)
    elif clock == 1:
        # select CLK2 input, power down CLK1 input (AD9512)
        SpiWrite(0x45, 0x02)
        # confirm write
        SpiWrite(0x5a, 1)
        # select IN0 (SY58017)
        WbSetBits('control', 3, 0)
    elif clock == 2:
        # select CLK2 input, power down CLK1 input (AD9512)
        SpiWrite(0x45, 0x02)
        # confirm write
        SpiWrite(0x5a, 1)
        # select IN1 (SY58017)
        WbSetBits('control', 3, 1)
    else:
        raise Exception("Unknown clock!")

# voltage in [V]
def SetTriggerThreshold(voltage):
    vdd = 5
    bits = 16
    code = int((voltage/vdd) * (2**bits))
    code &= (2**bits)-1
    WbWrite('trig_threshold', code)

# compensation <-1;1)
def SetVcxoFrequency(compensation):
    vdd = 3.3
    bits = 16
    voltage = vdd * (compensation + 1) / 2
    code = int((voltage/5) * (2**bits))
    code &= (2**bits)-1
    WbWrite('vcxo_voltage', code)

# channel: 1, 2, 0 (for both)
# memory: 'set', 'res', '*'
def ClearMemory(channel, memory):
    if channel == 1:
        address = 'ch1_'
    elif channel == 2:
        address = 'ch2_'
    elif channel == 0:
        ClearMemory(1, memory)
        ClearMemory(2, memory)
        return
    else:
        raise Exception("Bad channel!")
    
    if memory == "set":
        address += 'set_mem'
    elif memory == "res":
        address += 'res_mem'
    elif memory == "*":
        ClearMemory(channel, 'set')
        ClearMemory(channel, 'res')
        return
    else:
        raise Exception("Bad memory!")
    
    WbWriteMulti(address, (0,)*2048)

# channel: 1, 2, 0 (for both)
# pulse: 0, 1
# which: 0 - both, 1 - set, 2 - reset
def ConfigurePulse(channel, start, stop, pulse, which = 0):
    if channel == 1:
        address = 'ch1_'
    elif channel == 2:
        address = 'ch2_'
    elif channel == 0:
        ConfigurePulse(1, start, stop, pulse)
        ConfigurePulse(2, start, stop, pulse)
        return
    else:
        raise Exception("Bad channel!")
    
    if (start < 0) or (start > 65535):
        raise Exception("Bad start!")
    if (stop < 0) or (stop > 65535):
        raise Exception("Bad stop!")
    if (stop < start):
        raise Exception("Bad order!")
    
    pulse &= 1
    
    if which < 2:
        position = start/32
        shift = start%32
        WbSetBitsMulti(address+'set_mem', ((1<<shift),), ((pulse<<shift),), start=position, to=position+1)

    if which != 1:
        position = stop/32
        shift = stop%32
        WbSetBitsMulti(address+'res_mem', ((1<<shift),), ((pulse<<shift),), start=position, to=position+1)

def CreatePulse(channel, start, stop, which = 0):
    ConfigurePulse(channel, start, stop, 1, which)

def DeletePulse(channel, start, stop):
    ConfigurePulse(channel, start, stop, 0)

def PrintBit(value, bit, label, ifSet, ifNotSet):
    print('  '+'['+str(bit)+']'+label+':')
    if value & (1<<bit):
        print(ConColors.GREEN+'    1: '+ifSet+ConColors.NONE)
    else:
        print(ConColors.RED+'    0: '+ifNotSet+ConColors.NONE)

def Status():
    print('Status register:')
    status = WbRead('status')
    PrintBit(status, 0, 'Clock infrastructure configuration', 'busy', 'idle')
    PrintBit(status, 1, 'VCXO DAC', 'busy', 'idle')
    PrintBit(status, 2, 'Trigger threshold DAC', 'busy', 'idle')
    PrintBit(status, 3, 'Clock selection', 'busy', 'idle')
    PrintBit(status, 4, 'Delay configuration', 'busy', 'idle')
    PrintBit(status, 5, 'Channel 1 output', 'enabled', 'disabled')
    PrintBit(status, 6, 'Channel 2 output', 'enabled', 'disabled')
    PrintBit(status, 7, 'Channel 1', 'running', 'stopped')
    PrintBit(status, 8, 'Channel 2', 'running', 'stopped')

def Control():
    print('Control register: ')
    control = WbRead('control')
    PrintBit(control, 0, 'Clock mux', 'VCXO', 'FPGA LOOP')
    PrintBit(control, 2, 'Channel 1 output', 'enabled', 'disabled')
    PrintBit(control, 3, 'Channel 2 output', 'enabled', 'disabled')
    PrintBit(control, 4, 'Channel 1 mode', 'continuous', 'stop/one-shot')
    PrintBit(control, 6, 'Channel 2 mode', 'continuous', 'stop/one-shot')
    PrintBit(control, 8, 'LEDs', 'blinking', 'normal operation')
    PrintBit(control, 9, 'AD9512', 'in reset', 'running')

###################################################################
## Test functions
###################################################################

def Init():
    WbWrite('control', 0)
    SpiInit()
    Ad9512Init()
    SetTriggerThreshold(0.5)
    SelectClock(0)
    SetVcxoFrequency(-0.2105) # 125.0000 MHz
    ClearMemory(0, '*')

def TestPulse(channel = 1):
    # configure channel
    WbWrite('overflow', 17820)
    WbWrite('trig_latency', 8)

    start = 5
    length = 1
    stop = start + length
    CreatePulse(channel, start + 2*channel, stop + 2*channel)
    # CreatePulse(channel, 2, 2, 1)

    # CreatePulse(channel, 5, 10)
    # CreatePulse(channel, 25, 30)
    # CreatePulse(channel, 125, 130)

    WbWrite('ch'+str(channel)+'_delay_set', 1)
    WbWrite('ch'+str(channel)+'_delay_res', 100)
    # start channel
    channel -= 1
    channel %= 2
    WbSetBits('control', 0x4<<channel, 0x4<<channel) # set output enable
    channel *= 2
    WbSetBits('control', 0x30<<channel, 0x10<<channel) # set mode CONT.

Init()
Control()
Status()
TestPulse(1)
TestPulse(2)
Control()
Status()
